//
//							PHASOR
// Visualization program for seismic data stored in SEG-Y format 
// http://www.seg.org/documents/10161/77915/seg_y_rev1.pdf
// Performs phase rotation to the data between -180 to +180 degrees
// and allows to monitor the rotated data in a separate window
//
// Copyright 2011-2015 by Yuri Ganshin, CMI, University of Wyoming, Laramie.
// http://www.uwyo.edu/cmi/
//
// This program is free sofware released using the Fast Light Tool Kit (FLTK).
// For conditions of distribution and use, see the license agreement in 
//     http://www.fltk.org/COPYING.php
// As a general rule, "free software" means that the user have the freedom
// to run, copy, distribute, study, modify and improve the software.
//
// Please report all bugs and problems on the following page:
// http://www.uwyo.edu/cmi/
//     
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <cfloat>  // declares FLT_MAX
//
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>
#include <FL/fl_draw.H>	
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_Input_Choice.H>
#include <FL/fl_show_colormap.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>
#include <FL/fl_message.H>
#include <FL/fl_ask.H> // for fl_alert() and fl_message()
#include <math.h>
#include <fftw3.h>
/**
 The standard FFTW distribution works most efficiently and without any problem 
 for arrays whose size can be factored into small primes (2, 3, 5, and 7). 
 It was noticed that for the array lengths with larger-size prime factors (17 and 
 above), the complex transforms might occasionally produce incorrect results in 
 the form of high-frequency noise.    
 **/
using namespace std;

/**********************   Spectra-Related   *******************/
#define Re(z,i) ((z)[2*(i)])
#define Im(z,i) ((z)[2*(i)+1])

#define FOLD		13		// Multiplicity of spectral stacks
#define TWIN		1000	// Max. time window [msec] for spectral analysis
#define PALSZ		11
#define TIME_BAR	60
#define HDR_BAR		24
#define RIM			 8		// RIM == border for canvas
#define MAX_VAL		100		// Maximum number of trace headers to plot
//
#ifndef FLT_MAX
#define FLT_MAX 1e+37
#endif
//
#ifndef M_PI
#define M_PI 3.14159265359
#endif
//
////////////////////////////////////////////////////////////////
/****               List of GLOBAL variables               ****/
FILE *in_fp = NULL;
char inpname[200];
int nsamp, dsamp, Ntr, convert_mode=1;
short format,measure;
int tr_sta,tr_fin,samp_sta,samp_fin,hdr_step=5;
int tr_vis=-1,samp_vis, wiggle_trace=0, envelope=0;
int numpix=3;
int hdr_val[MAX_VAL];
//
int canvas_w = 760, canvas_h =960;
//
int word_no=5;			// CDP for header plot
unsigned int tr_size, samp_step=5;
float Norm, gain, elev_scal, coor_scal, *seis;
float *trace1, *trace2;
char *hdrs,c_word[4],cgbuf[120];
char formats[][44] =    // string formats[9]
{
	"",
	"Data Format:  4-byte IBM floating-point",     // 1
	"Data Format:  4-byte integer",
	"Data Format:  2-byte integer", 
	"Data Format:  4-byte fixed-point with gain",
	"Data Format:  4-byte IEEE floating-point",    // 5
	"Data Format:  byte-swapped IEEE f.p.",        // 6
	"Data Format:  non-standard 4-byte IEEE f.p.",  
	"Data Format:  1-byte integer"
};
char text_hdr[3201];
//
//  Rotation Related
float rotangle=0.0;
//
////////////////////////////////////////////////////////////////
/****            List of Function Prototypes               ****/

void read_file_hdr(const char *name);
void read_seis_dat();
void convert2(char* input);
void convert4(char* input);
void convert_hdrs(char* array);
float ibm2ieee(const float ibm);
float efloat(float value);
unsigned char to_ascii ( unsigned char c );
float conv3(float *dat,int NS1,int NS2,int index);

////////////////////////////////////////////////////////////////
/****         Global pointers to the GUI objects           ****/
////////////////////////////////////////////////////////////////
Fl_Double_Window *seis_window, *proc_window;
Fl_Window *opts_window, *hdr_window;
Fl_Int_Input	*tr_sta_field, *tr_fin_field;
Fl_Int_Input	*t_sta_field,  *t_fin_field;
Fl_Float_Input	*gain_field;
Fl_Button *apply, *next_tr, *add_wiggle, *show_hdr, *end_this;
Fl_Button *apply_rotate, *add_envelope;
Fl_Value_Slider *rotval;

////////////////////////////////////////////////////////////////
/****        List of Widgets and Drawing functions         ****/
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
/********             File-choosing MENU               ********/
////////////////////////////////////////////////////////////////
class File_menu : public Fl_Window {
	Fl_Native_File_Chooser *fc;
	
	void read_data(const char *filename)	// Process the chosen file
	{
		strcpy(inpname,filename);
		tr_vis=1;			// will wait while(tr_vis<0)
	}
	
	// Handle an 'Open' request from the menu
	static void open_cb(Fl_Widget *w, void *v) {
		File_menu *choice = (File_menu*)v;
		choice->fc->title("Open");
		choice->fc->type(Fl_Native_File_Chooser::BROWSE_FILE);
		switch ( choice->fc->show() ) {
			case -1: break;	// Error
			case  1: break; 	// Cancel
			default:		// choice->fc->show() returns 0
				choice->read_data(choice->fc->filename());
				// choice->deactivate();				
				break;
		}
	}

	static void quit_cb(Fl_Widget *w, void *v) {
		exit(0);
	}
	
public:
	File_menu() : Fl_Window(500,300,320,160,"Choose the file") {
		Fl_Menu_Bar *menu = new Fl_Menu_Bar(0,0,300,25);
		menu->add("&File/&Open",  FL_COMMAND+'o', open_cb, (void*)this);
		menu->add("&File/&Quit",  FL_COMMAND+'q', quit_cb);
		
		// Describe the chooser
		Fl_Box *box = new Fl_Box(20,45,w()-40,h()-65);
		box->color(FL_GRAY-1); 
		box->box(FL_FRAME_BOX);
		box->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE|FL_ALIGN_WRAP);
		box->label("Choose seismic binary SEG-Y file\n\
with 4-byte floating point data format\n\
[ IBM, IEEE, or native byte order ]");
		box->labelsize(16);
		// Initialize the file chooser
		fc = new Fl_Native_File_Chooser();
		fc->filter("SEG-Y Binary\t*.sgy\n");			 
		end();
	}
};

////////////////////////////////////////////////////////////////////////////
/*****************        Drawing Seismic Data          *******************/
//   Double-buffered window eliminates flickering during updates
////////////////////////////////////////////////////////////////////////////
class Draw_seis : public Fl_Double_Window 
{

public:
  	Draw_seis(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
	void draw(void)  // Function draw() within class Draw_seis
	{      /**   don't declare reserved 'x' and 'y' variables !  **/
		Fl_Double_Window::draw();
		unsigned i, j;
		int xx, xx1=0, yy, x_dim, y_dim, *hdr4;
		int val, tr1, tr2, sta_pix;
		short *hdr2;
		float yi, a1, a2;
		double dx,dtr;
		
		if(samp_vis<50) samp_step=5;		// Step for Time Labels
		else if(samp_vis<500) samp_step=25;
		else if(samp_vis<1000) samp_step=50;	
		else if(samp_vis<2000) samp_step=100;
		else if(samp_vis<5000) samp_step=250;
		else if(samp_vis<10000) samp_step=500;
		else samp_step=1000;
		
		hdr2=(short*)hdrs;	// To get trace ID (0=Unknown,1=Seismic,2=Dead)
		hdr4=(int*)hdrs;
		for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++) 
		{
			val = hdr4[tr1*60+word_no];
			if(word_no>=10 && word_no<=14) val=(int)((float)val*elev_scal);
			if(word_no>=18 && word_no<=46) val=(int)((float)val*coor_scal);
			if(j<MAX_VAL) hdr_val[j]=val;
		}			
		
		/**    Estimate X-shft from the number of numerals	  **/
		tr1=tr_vis/2;      // take middle trace as estimate
		val = hdr4[tr1*60+word_no];  // 60 = number_of_4-byte_words
		if(word_no>=10 && word_no<=14) val=(int)((float)val*elev_scal);
		if(word_no>=18 && word_no<=46) val=(int)((float)val*coor_scal);
		if(val<10) numpix=3;
		else if(val<100) numpix=7;
		else if(val<1000) numpix=11;
		else if(val<10000) numpix=15;		
		else if(val<100000) numpix=19;
		else if(val<1000000) numpix=23;
		else numpix=27;			//  numpix == X-shift in pixels
		
		x_dim = w() - TIME_BAR;  y_dim = h() - HDR_BAR;
		//cout << "Xdim = " << x_dim << ", Y_dim = " << y_dim << endl;	
		/////////////////////////////////////////////////////////////////

		fl_rectf(0,0,w(),h(),  0,  0,  0);			// Black bounding rectangle
		fl_rectf(1,1,w()-2,h()-2,230,230,230);		// Gray bounding rectangle
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(TIME_BAR,HDR_BAR,x_dim-1,y_dim-1,255,255,255);
		fl_color(FL_BLACK);	fl_font(0,14);	fl_line_style(0,1); // solid, 1 pixel wide
		
		dtr = (double)x_dim/tr_vis;  //  decimal  pixels/trace
		xx1 = sta_pix = TIME_BAR + (int)(dtr/2.0) + 1;
		
		///////////////////////////////////////////////////////////////////////
		// Variable Area Display utilizing cubic convolution for resampling	 //
		// and linear interpolataion for filling inter-trace space			 //
		///////////////////////////////////////////////////////////////////////
		for(tr1=0;tr1<tr_vis;tr1++)
		{										
			i = tr1*x_dim/tr_vis;
			xx = i + sta_pix;
			tr2 = tr1+1; 
			dx = fabs((double)i - dtr*tr1)/dtr;	//   0.0 <= dx <= 1.0

			if(hdr2[tr1*120+14]==2) for(i=0;i<nsamp;i++) trace1[i]=0.0; // avoid NAN's
			else memcpy(trace1, seis+tr1*nsamp, nsamp * sizeof(float));

			if(hdr2[tr2*120+14]==2) for(i=0;i<nsamp;i++) trace2[i]=0.0; // avoid NAN's
			else memcpy(trace2, seis+tr2*nsamp, nsamp * sizeof(float));

			for(j=0;j<y_dim;j++)  // possible: j+=3 for fl_yxline() 
			{
				yy = j+HDR_BAR;	// yy = 24 (j==0)
				// Resample data for visualization using Cubic Convolution Method 
				a1=conv3(trace1+samp_sta,samp_vis,y_dim,j);	
				a2=conv3(trace2+samp_sta,samp_vis,y_dim,j);
				yi = a1+(a2-a1)*dx; // Inter-trace linear interpol: a1<=yi<=a2
				val = (int)((yi*gain*15)/Norm);
				if(val>0)fl_xyline(xx, yy, xx+val); // horiz line for VA display
				if(wiggle_trace) if(j>0) fl_line(xx+val,yy,xx1,yy-1); // overlay
				xx1 = xx+val;						
			}								
		}		
		
		for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++)	////  Draw Trace Headers 
		{
			i = tr1*x_dim/tr_vis;
			xx = i + sta_pix;
			fl_yxline(xx, HDR_BAR, HDR_BAR-6);		 
			if(j<MAX_VAL) sprintf(cgbuf,"%d",hdr_val[j]); 
			else sprintf(cgbuf,"%d",99999);
			fl_draw(cgbuf,xx-numpix,HDR_BAR-10);	
		}											////  Draw Trace Headers
		
		/////////////////////////////////////////////////  Draw Time labels
		sprintf(cgbuf,"Time ");  fl_draw(cgbuf,10,HDR_BAR+8);
		sprintf(cgbuf,"[ ms ]");  fl_draw(cgbuf,10,HDR_BAR+26);	
		for(i=0;i<samp_vis;i+=samp_step)
		{
			yy = y_dim*i/samp_vis + HDR_BAR;
			if(i>0) 
			{
				val=(samp_sta+i)*dsamp/1000;
				sprintf(cgbuf,"%5d",val); fl_draw(cgbuf,10,yy+4); // text string
				fl_xyline(TIME_BAR-10,yy,TIME_BAR+x_dim);  // horiz line			
			}
		}
		
	} 	// void draw(void)	
	
} *seismic;

/**
 Hilbert transform allows to manipulate Fourier phase of a signal
 (changing its structure) without affecting its amplitude spectrum.
 Conventional Hilbert Transform is equivalent to a rotation by +90˚
 =================================================================
 Let G(f)=(a + bi); is a complex signal that we want to transform.
 Hilbert operator H(f) is defined in frequency domain as:
 -i  if f>0 
  0  if f=0
  i  if f<0 
 where i = sqrt(-1);	
 By multiplying these complex numbers we'll get:
 H(f)*G(f) = -i*(a + bi) =  b - ia			   if f>0 
 and H(f)*G(f) =  i*(a + bi) = -b + ia			if f<0
 Thus, for positive frequencies:
 Re(Hilb(G(f))) = Im(G(f));		Im(Hilb(G(f))) = -Re(G(f));
 and for negative frequencies:
 Re(Hilb(G(f))) = -Im(G(f));	Im(Hilb(G(f))) = Re(G(f));
 and for zero frequency (DC):
 Re(Hilb(G(f))) = 	Im(Hilb(G(f))) = 0.0;
 ================================================================
 In general case, to produce a phase shift to any angle a,
 Hilbert operator H(f) is defined as:
 e^(-ia) = cos(a)-i*sin(a);  if f>0 
 0							 if f=0
 e^(+ia) = cos(a)+i*sin(a);  if f<0
 After some math:
 Re(Hilb(G(f),a)) = cos(a)*Re(G(f)) + sin(a)*Im(G(f));		if f>0
 Im(Hilb(G(f),a)) = cos(a)*Im(G(f)) - sin(a)*Re(G(f));		if f>0
 Re(Hilb(G(f),a)) = cos(a)*Re(G(f)) - sin(a)*Im(G(f));		if f<0
 Im(Hilb(G(f),a)) = cos(a)*Im(G(f)) + sin(a)*Re(G(f));		if f<0
 
 Info:  cos(90˚) = cos(-90˚) = 0.0;  sin(90˚)=1; sin(-90˚)=-1;		
**/
/////////////////////////////////////////////////////////////////////
/**************       Drawing Processed Data        ****************/
/////////////////////////////////////////////////////////////////////
class ProcCanvas : public Fl_Widget 
{
public:
	ProcCanvas(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}

	void draw()
	{
		unsigned i, j, tr, pos;
		int xx, xx1=0, yy, x_dim, y_dim;
		int val, t, tr1, tr2, sta_pix;
		short *hdr2;
		float yi, a1, a2, *proc, *envel;
		double dx,dtr,val1,val2;
		// Fast Fourier Transform related:
		fftw_complex *in,*out,*out3;
		fftw_plan fwd,bck,bck3;
		double *tmp1,*tmp2,*tmp3,sina,cosa;

		x_dim = w() - TIME_BAR;  y_dim = h() - HDR_BAR;
		/////////////////////////      !!!!!!     //////////////////////////////
		fl_rectf(x()-1,y()-1,w()+2,h()+2,  0,  0,  0);	// Black bounding rectangle
		fl_rectf(x(),  y(),  w(),  h(),  230,230,230);	// Gray bounding rectangle
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(x()+TIME_BAR,y()+HDR_BAR,x_dim,y_dim,255,255,255);
		fl_color(FL_BLACK);	fl_font(0,14);	 // Helvetica regular, 14 pixel
		fl_line_style(0,1); // solid, 1 pixel wide
		dtr = (double)x_dim/tr_vis;  //  decimal  pixels/trace
		xx1 = sta_pix = TIME_BAR + (int)(dtr/2.0) + 1;
		hdr2=(short*)hdrs;	// To get trace ID (0=Unknown,1=Seismic,2=Dead)
		
		//////////////////////////////////////////////////////////////////////
		////    Processing    Processing    Processing    Processing	  ////
		//////////////////////////////////////////////////////////////////////

		//  sizeof(fftw_complex) = 16 bytes
		in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * nsamp);
		out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * nsamp);
		out3 = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * nsamp);		
		// Forward and backward in-place transforms:		
		fwd = fftw_plan_dft_1d(nsamp,in,in,-1,FFTW_ESTIMATE); 
		bck = fftw_plan_dft_1d(nsamp,out,out,1,FFTW_ESTIMATE);
		bck3 = fftw_plan_dft_1d(nsamp,out3,out3,1,FFTW_ESTIMATE);
				
		tmp1=(double *)in;	// To translate complex-number type into double		
		tmp2=(double *)out;
		tmp3=(double *)out3;	

		proc = new (nothrow) float[(tr_vis+1) * nsamp];		
		envel = new (nothrow) float[(tr_vis+1) * nsamp];
		if(envel == 0)
		{ fl_message("ERROR: Not enough memory to process data!\n\
						  Close the window to exit...");	   exit(1); }

		sina = sin(rotangle*M_PI/180.0);	// angle must be in radians
		cosa = cos(rotangle*M_PI/180.0);	// angle must be in radians

		for(tr=0;tr<tr_vis;tr++)	////     PHASE ROTATION
		{
			if(hdr2[tr*120+14]==2) for(i=0;i<nsamp;i++) trace1[i]=0.0; // avoid NAN's
			else memcpy(trace1, seis+tr*nsamp, nsamp * sizeof(float));
					
			//  Read seismic data into complex array and transform it
			for(i=0;i<nsamp;i++) 
			{ Re(tmp1,i)=(double)trace1[i]; Im(tmp1,i)=0.0; }
			fftw_execute(fwd);		// Forward in-place transform with in[]
			
			for(i=1;i<=nsamp/2;i++)
			{				
				Re(tmp2,i) = cosa*Re(tmp1,i) + sina*Im(tmp1,i);
				Im(tmp2,i) = cosa*Im(tmp1,i) - sina*Re(tmp1,i);
            Re(tmp3,i) = -Im(tmp1,i);
            Im(tmp3,i) = Re(tmp1,i);				
				// negative Frequency
				if(i==nsamp/2) break;// do not duplicate Nyqusist frequency
				j=nsamp-i;	//  Backwards order for 'negative' frequencies
				Re(tmp2,j) = cosa*Re(tmp1,j) - sina*Im(tmp1,j);
				Im(tmp2,j) = cosa*Im(tmp1,j) + sina*Re(tmp1,j);
            Re(tmp3,j) = Im(tmp1,j);
            Im(tmp3,j) = -Re(tmp1,j);							
			}
			Re(tmp2,0) = Im(tmp2,0) = 0.0;   Re(tmp3,0) = Im(tmp3,0) = 0.0;
			
			fftw_execute(bck);		// Backward in-place transform with out[]
			fftw_execute(bck3);
			
			for(i=0;i<nsamp;i++)	// Fill-in proc[] and normalize with 'nsamp'
			{
				proc[tr*nsamp+i]=(float)(Re(tmp2,i)/nsamp);
				val1 = (double)trace1[i];	// Input seismic trace
				val2 = Re(tmp3,i)/nsamp;	// Quadrature trace [Hilbert transformed]	
				envel[tr*nsamp+i] = (float)sqrt(val1*val1 + val2*val2);				
			}
		}
		///////////////////////////////////////////////////////////////////////
		//     Visualization of Rotated Data and Instantaneous Amplitude     //
		//                                                                   //
		//  Variable Area Display utilizing cubic convolution for resampling //
		//  and linear interpolataion for filling inter-trace space          //
		///////////////////////////////////////////////////////////////////////
		for(tr1=0;tr1<tr_vis;tr1++)	// Dispaly all rotated traces
		{		
			i = tr1*x_dim/tr_vis;
			xx = x() + i + sta_pix;
			dx = fabs((double)i - dtr*tr1)/dtr;	//   0.0 <= dx <= 1.0			
			tr2 = tr1+1;

			memcpy(trace1, proc+tr1*nsamp, nsamp * sizeof(float));
			memcpy(trace2, proc+tr2*nsamp, nsamp * sizeof(float));

			fl_color(FL_BLACK);	fl_line_style(0,1);		
			for(j=0;j<y_dim;j++)
			{
				yy = y()+j+HDR_BAR;
				a1=conv3(trace1+samp_sta,samp_vis,y_dim,j);
				a2=conv3(trace2+samp_sta,samp_vis,y_dim,j);
				yi = a1+(a2-a1)*dx; // Inter-trace linear interpol: a1<=yi<=a2
				val = (int)((yi*gain*15)/Norm);
				if(val>0)fl_xyline(xx, yy, xx+val); // horiz line for VA display
				if(wiggle_trace) if(j>0) fl_line(xx+val,yy,xx1,yy-1); // overlay
				xx1 = xx+val;						
			}
			if(envelope==1)	// Optionally dispaly trace envelope
			{
				memcpy(trace1, envel+tr1*nsamp, nsamp * sizeof(float));
				memcpy(trace2, envel+tr2*nsamp, nsamp * sizeof(float));

				fl_color(FL_RED);	fl_line_style(0,2);		
				for(j=0;j<y_dim;j++) 
				{
					yy = y()+j+HDR_BAR;
					a1=conv3(trace1+samp_sta,samp_vis,y_dim,j);
					a2=conv3(trace2+samp_sta,samp_vis,y_dim,j);
					yi = a1+(a2-a1)*dx; // Inter-trace linear interpol: a1<=yi<=a2
					val = (int)((yi*gain*15)/Norm);
					if(j>0) fl_line(xx+val,yy,xx1,yy-1); // Wiggle overlay only
					xx1 = xx+val;						
				}
			}
		}		
		
		fl_color(FL_BLACK);
		for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++)	////  Draw Trace Headers
		{
			i = tr1*x_dim/tr_vis;
			xx = x() + i + sta_pix;
			fl_yxline(xx, y()+HDR_BAR, y()+HDR_BAR-6);		 
			if(j<MAX_VAL) sprintf(cgbuf,"%d",hdr_val[j]); 
			else sprintf(cgbuf,"%d",99999);
			fl_draw(cgbuf,xx-numpix,y()+HDR_BAR-10);	
		}			
		
		xx = x()+10;  yy = y()+HDR_BAR;					////  Draw Time labels
		sprintf(cgbuf,"Time ");  fl_draw(cgbuf,xx,yy+8);
		sprintf(cgbuf,"[ ms ]"); fl_draw(cgbuf,xx,yy+26);
				
		xx = x()+TIME_BAR;
		for(i=0;i<samp_vis;i+=samp_step)
		{
			yy = y_dim*i/samp_vis + y() + HDR_BAR;
			if(i>0) 
			{
				val=(samp_sta+i)*dsamp/1000;
				sprintf(cgbuf,"%5d",val); fl_draw(cgbuf,x()+10,yy+4);
				fl_xyline(xx-10, yy, xx+x_dim);	  // horiz line	
			}
		}
		
		fftw_destroy_plan(fwd); fftw_destroy_plan(bck); fftw_destroy_plan(bck3);
		fftw_free(in); fftw_free(out); fftw_free(out3); delete[] proc; delete[] envel;
		
	}	// void draw()
	
} *proc_draw;
//
//
////////////////////////////////////////////////////////////////
/************    List of Call Back functions    ***************/
////////////////////////////////////////////////////////////////

void input_tr_sta(Fl_Widget *, void *) {
	tr_sta = atoi(tr_sta_field->value());
}

void input_tr_fin(Fl_Widget *, void *) {
	tr_fin = atoi(tr_fin_field->value());
}

void input_t_sta(Fl_Widget *, void *) {
	samp_sta = atoi(t_sta_field->value())*1000/dsamp;
}

void input_t_fin(Fl_Widget *, void *) {
	samp_fin = atoi(t_fin_field->value())*1000/dsamp;
}

void input_gain(Fl_Widget *, void *) {
	gain = atof(gain_field->value());
}

void spinner_cb(Fl_Widget *w, void *) {
   Fl_Spinner *s = (Fl_Spinner *)w;	
   hdr_step = (int)s->value();
	// printf("Spinner Value = %d\n",hdr_step);
}

void input_choice_cb(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_hdr=(Fl_Input_Choice *)data;
	
	strncpy(c_word,in_hdr->value(),3);
	sscanf(c_word,"%d", &word_no);
}

void show_hdr_cb(Fl_Widget *, void *) 
{
   // fl_message(text_hdr); -> accomdates only 12 rows out of 40
	hdr_window = new Fl_Window(676,676,"File's Textual header");

	Fl_Box* txt = new Fl_Box(FL_FRAME_BOX,8,8,660,660,text_hdr);
	txt->color(FL_WHITE);	
	txt->labelfont(0);
	txt->labelsize(14);
	
	hdr_window->resizable(txt); 
	hdr_window->end();
	hdr_window->show();
}

void exit_b(Fl_Widget *, void *) {
	exit(0);
}
                       
//////////   Rotation Angle Slider
static void slide_input(Fl_Widget* rotval) 
{
    rotangle = ((Fl_Valuator*)rotval)->value();
}

// Redraw display with updated data
void doRedraw(Fl_Widget *, void*)
{
	if(tr_fin>Ntr) tr_fin=Ntr;	
	if(tr_sta>=tr_fin) tr_sta=1; 
	tr_vis = tr_fin-tr_sta+1; // if trace count starts from '1'

	if(samp_fin>nsamp) samp_fin=nsamp;
	if(samp_sta>=samp_fin) samp_sta=0;
	samp_vis = samp_fin-samp_sta;
		
	read_seis_dat();
	seismic->redraw();
}

void doNext(Fl_Widget*, void*)
{
	if(tr_fin<Ntr)
	{
		tr_fin += tr_vis;			// tr_vis remains unchanged
		if(tr_fin>Ntr) tr_fin=Ntr;
		tr_sta = tr_fin-tr_vis+1;	// from tr_vis formula

		sprintf(cgbuf,"%d",tr_sta);  tr_sta_field->value(cgbuf);
		sprintf(cgbuf,"%d",tr_fin);  tr_fin_field->value(cgbuf);
			
		read_seis_dat();
		seismic->redraw();
	}
}

void do_wiggle(Fl_Widget*, void*)
{
	if(wiggle_trace==0) wiggle_trace=1;
	else wiggle_trace=0; 
	seismic->redraw();
}

////////////////////////////////////////////////////////////////////////
///////////////     Window with Phase Rotated data      ////////////////
////////////////////////////////////////////////////////////////////////
void Process(Fl_Widget *, void*)
{  

	sprintf(cgbuf,"Phase Rotation Angle = %.0f degrees",rotangle);
	proc_window = new Fl_Double_Window(canvas_w+RIM*2,canvas_h+RIM*2,cgbuf);
			
	proc_draw = new ProcCanvas(RIM,RIM,canvas_w,canvas_h);
	//seismic->redraw();	

	proc_window->end();
	proc_window->resizable(proc_window); // Critical !!!
	proc_window->show();
	add_envelope->activate();
}
/////////////////   Optionally  Add the Envelope (Inst. Ampl.) 
void do_envelope(Fl_Widget *, void*)
{  

	if(envelope==0) envelope=1;
	else envelope=0; 
	proc_draw->redraw();

}
//
//
/////////////////////////////////////////////////////////////////////
/****   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   ****/
/////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	char cbuf[90],cbuf0[120],cbuf1[99],cbuf2[99],cbuf3[99],cbuf4[99];
	/////////////////////////////////////////////////////////////////
	
	Fl::scheme(NULL); //  "plastic" for Aqua and "gtk+"
	// NULL use the scheme defined in the FLTK_SCHEME env.var. 
	
	if(argc == 2) strcpy(inpname,argv[1]);		// Do not use terminal
	else	// if you want to create an Apple  *.app bundle application
	{
		File_menu *choice = new File_menu();
		choice->show();
		choice->end(); // no more widgets will be added to the window
		while(tr_vis<0) Fl::wait();
		choice->hide();
	}

	read_file_hdr(inpname);
	read_seis_dat();			// into new seis[]

	///////////////////////////////////////////////////////////
	strcpy(cbuf0,"Seismic Data");
	seis_window = new Fl_Double_Window(canvas_w+RIM*2,canvas_h+RIM*2,cbuf0);
	
	seismic = new Draw_seis(RIM,RIM,canvas_w,canvas_h);
	// seismic->end();	
	
	seis_window->end();
	seis_window->resizable(seis_window); // Critical !!!
	seis_window->show();
	///////////////////////////////////////////////////////////
		
	opts_window = new Fl_Window(1050,50,420,875,"Display Manager");
	// The follwing widgets belong to this window
	
	Fl_Box* label1 = new Fl_Box(FL_FRAME_BOX,20,4,380,122,""); label1->color(6);
	
	Fl_Box* label2 = new Fl_Box(80,8,250,20,"  File Information  ");
	label2->labelcolor(FL_BLACK);
	label2->labelfont(FL_HELVETICA_BOLD);	label2->labelsize(16);
	
	sprintf(cbuf1,"Number of traces = %d,  %d samples each",Ntr,nsamp);
	Fl_Box* label3 = new Fl_Box(80,28,250,20,cbuf1);
	label3->labelfont(FL_HELVETICA);	label3->labelsize(14);
	
	sprintf(cbuf2,"Sample interval = %d microseconds",dsamp);
	Fl_Box* label4 = new Fl_Box(80,46,250,20,cbuf2);
	
	sprintf(cbuf3,"Trace length = %d milliseconds [ms]",nsamp*dsamp/1000);
	Fl_Box* label5 = new Fl_Box(80,64,250,20,cbuf3);
	
	Fl_Box* label6 = new Fl_Box(80,82,250,20,formats[format]);
	
   if(measure<1 || measure>2) sprintf(cbuf4,"Measurement system - unknown");
   if(measure==1) sprintf(cbuf4,"Measurement system - metric [meters]");
   if(measure==2) sprintf(cbuf4,"Measurement system - imperial [feet]");	
	Fl_Box* label7 = new Fl_Box(80,100,250,20,cbuf4);		

	/////////////////////////////////////////////////////////////////				
	show_hdr = new Fl_Button(135,132,160,25,"Show file's text header");
	show_hdr->color(134);			show_hdr->labelfont(2);			
	show_hdr->labelsize(14);		show_hdr->box(FL_RSHADOW_BOX);	
	show_hdr->callback(show_hdr_cb);	
	
	/////////////////////////////////////////////////////////////////	

	Fl_Box* label_tr = new Fl_Box(80,150,140,50,"Range of displayed traces :");
	label_tr->labelfont(FL_HELVETICA_BOLD);
	label_tr->labelsize(15);
	
	tr_sta_field = new Fl_Int_Input(80, 190, 70, 25, " from ");
	sprintf(cbuf,"%d",tr_sta);  tr_sta_field->value(cbuf);
	tr_sta_field->callback(input_tr_sta);
	tr_fin_field = new Fl_Int_Input(180, 190, 70, 25, " to ");
	sprintf(cbuf,"%d",tr_fin);  tr_fin_field->value(cbuf);
	tr_fin_field->callback(input_tr_fin);
	
	next_tr = new Fl_Button(285, 190, 100, 25, " Next traces");
	next_tr->color(62);
	next_tr->labelfont(1);	next_tr->box(FL_RSHADOW_BOX);	
	next_tr->callback(doNext);

	Fl_Box* label_t = new Fl_Box(80,210,140,50,"Time range [ms] to display:");
	label_t->labelfont(FL_HELVETICA_BOLD);
	label_t->labelsize(15);	
	
	t_sta_field = new Fl_Int_Input(80, 250, 70, 25, " from ");
	sprintf(cbuf,"%d",samp_sta*dsamp/1000); t_sta_field->value(cbuf); 
	t_sta_field->callback(input_t_sta);
	t_fin_field = new Fl_Int_Input(180, 250, 70, 25, " to ");
	sprintf(cbuf,"%d",samp_fin*dsamp/1000); t_fin_field->value(cbuf); 
	t_fin_field->callback(input_t_fin);
	
	gain_field = new Fl_Float_Input(310, 250, 70, 25, "Gain: ");
	sprintf(cbuf,"%5.2f",gain);  gain_field->value(cbuf);
	gain_field->labelsize(14);
	gain_field->labelfont(FL_HELVETICA_BOLD);	
	gain_field->callback(input_gain);
	
	Fl_Box* label_hdr = new Fl_Box(50,280,300,54,
								"Select header plot parameter\n\
from the list of SEGY trace headers:");
	label_hdr->labelfont(FL_HELVETICA_BOLD);
	label_hdr->labelsize(15);
	Fl_Input_Choice in_hdr(50,330,280,25,"");
	in_hdr.callback(input_choice_cb, (void*)&in_hdr);
	
	in_hdr.add("  2 - Field record number");
	in_hdr.add("  3 - Trace number within Field record");
	in_hdr.add("  4 - Energy source point number"); 
	in_hdr.add("  5 - Ensemble number (CDP, CRT, ...)");
	in_hdr.add("  6 - Trace number within the ensemble");
	in_hdr.add("  9 - Source-receiver offset");
	in_hdr.add(" 10 - Receiver group elevation");  
	in_hdr.add(" 11 - Surface elevation at source");
	in_hdr.add(" 13 - Datum elevation at receiver");
	in_hdr.add(" 14 - Datum elevation at source");
	in_hdr.add(" 18 - Source coordinate X");
	in_hdr.add(" 19 - Source coordinate Y");
	in_hdr.add(" 20 - Group coordinate X");
	in_hdr.add(" 21 - Group coordinate Y");
	in_hdr.add(" 45 - ensemble (CMP) coordinate X");
	in_hdr.add(" 46 - ensemble (CMP) coordinate Y");
	in_hdr.add(" 47 - In-line number for 3-D data");
	in_hdr.add(" 48 - Cross-line number for 3-D data");
	in_hdr.add(" 49 - Shotpoint number (for 2-D data)");		
	in_hdr.value(3);  // CDP - default header plot parameter
	
	Fl_Box* comment = new Fl_Box(50,350,300,50,
        "the leading number indicates 4-byte integer word,\n\
starting from 0, within the 240-byte trace header");
	comment->labelfont(FL_HELVETICA);
	comment->labelsize(14);
	
	Fl_Spinner* spin = new Fl_Spinner(250, 410, 70, 25, 
                    "Header spacing, traces ");
	spin->labelsize(15);
	spin->labelfont(FL_HELVETICA_BOLD);	
	spin->minimum(1);
	spin->maximum(250);
	spin->step(5);
	spin->value(hdr_step);	
	spin->align(Fl_Align(FL_ALIGN_LEFT));
	spin->callback(spinner_cb);
	////////////////////////////////////////////////////////////////////////	
	
	apply = new Fl_Button(50,450,140,30,
								 "Redraw traces");
	apply->color(62);		apply->labelcolor(FL_BLACK);
	apply->labelfont(3);	apply->labelsize(16);	
	apply->box(FL_RSHADOW_BOX);	apply->callback(doRedraw);	
	
	
	add_wiggle = new Fl_Button(200, 453, 165, 25, 
										"Add/remove wiggles");
	add_wiggle->color(62);			add_wiggle->labelcolor(FL_BLACK);
	add_wiggle->labelfont(3);		add_wiggle->labelsize(14);										
	add_wiggle->box(FL_RSHADOW_BOX);	add_wiggle->callback(do_wiggle);
	
	////////////////////////////////////////////////////////////////////////
	////////////            Phase Rotation            //////////////////////	
	////////////////////////////////////////////////////////////////////////
	Fl_Box* label_ooo = new Fl_Box(FL_FRAME_BOX,20,500,380,24,
			"=======  Phase Rotation of the displayed data  =======");
	label_ooo->labelfont(0);  label_ooo->labelcolor(7);
	label_ooo->color(36);		label_ooo->labelsize(14);
                                 
	Fl_Box* label13 = new Fl_Box(40,525,100,50,"Phase Rotation:");
	label13->labelfont(FL_HELVETICA_BOLD); label13->labelsize(14);
	Fl_Box* label14 = new Fl_Box(330,525,60,50,"degrees");
	label14->labelfont(FL_HELVETICA_BOLD); label14->labelsize(14);
                                 
	rotval = new Fl_Value_Slider(150, 542, 175, 25);
	rotval->type(FL_HOR_SLIDER);
	rotval->bounds(-180, +180);     // Set up bounds of the slider
	rotval->value(0);					// Set up the default value
	rotval->step(15);	
	rotval->selection_color(5);
	rotval->labelsize(13); rotval->labelfont(FL_HELVETICA_ITALIC);
	rotval->callback(slide_input);	// Set up the callback function
	rotval->show();		    

	//////   Callback function to open Phase-Rotated window    /////////
	apply_rotate = new Fl_Button(35, 585, 140, 30, 
									 " Apply Rotation ");
	apply_rotate->color(62);				apply_rotate->labelsize(16); 
	apply_rotate->labelfont(3);			apply_rotate->labelcolor(FL_BLACK);          
	apply_rotate->box(FL_RSHADOW_BOX);	apply_rotate->callback(Process);	
		
	//////   Callback function to open Inst. Amplitude window    //////	

	add_envelope = new Fl_Button(185, 585, 195, 30,
										  " Add/remove envelope ");
	add_envelope->color(62);				add_envelope->labelsize(16); 
	add_envelope->labelfont(3);			add_envelope->labelcolor(FL_BLACK);          
	add_envelope->box(FL_RSHADOW_BOX);	add_envelope->callback(do_envelope);
	add_envelope->deactivate();
	////////////////////////////////////////////////////////////////////////			

	Fl_Box* label8 = new Fl_Box(FL_FRAME_BOX,8,640,405,170,""); 
	label8->color(6);
	Fl_Box* label9 = new Fl_Box(80,650,250,20,
							" Checking the phase of your seismic ");
	label9->labelcolor(FL_BLACK); label9->color(6); 
	label9->labelsize(17);  label9->labelfont(FL_HELVETICA_BOLD_ITALIC);
		
	Fl_Box* label10 = new Fl_Box(150,720,120,25,
"  To assess the phase of seismc waveform, first display a \n\
portion of data with a strong reflection in it.            \n\
In zero-phase data, we would expect peaks (or troughs)     \n\
in the trace to align with peaks in the envelope.          \n\
By scanning the phase rotation angle and watching the match\n\
of data peaks with the envelope, we can determine the angle\n\
that brings data to zero-phase.");
	label10->labelfont(FL_HELVETICA_ITALIC); label10->labelsize(14);

	///////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////
	end_this = new Fl_Button(165,825, 80, 30, "Exit");
	end_this->color(88);
	end_this->labelcolor(FL_BLACK);
	end_this->labelfont(3);	
	end_this->labelsize(18);	end_this->box(FL_RSHADOW_BOX);
	end_this->callback(exit_b);	
			
	opts_window->end(); 
	opts_window->show();

	return(Fl::run());
	
	fclose(in_fp);
   return 0;
}
/////////////////////////////////////////////////////////////////////
/****    END_MAIN   END_MAIN   ENF_MAIN   END_MAIN   END_MAIN   ****/
/////////////////////////////////////////////////////////////////////
/****               Called Function Definition                  ****/
/////////////////////////////////////////////////////////////////////
void read_file_hdr(const char *name)
{
	int i=1,j,N,tr;
	long file_size,offset;
	float amp,count,avr,sinc;
	double val,std;
	char array[3601];
	short *hdr2;	
	
	if( *( (char *)&i ) != 1 )	// TEST on COMPUTER ENDIANISM
	{
		fl_message("This application is supposed to run on\n\
Intel-based PC's with little-Endian data format.\n\
This is big-Endian computer. Close the window to exit");
		exit(1);
	}
	
	if ((in_fp = fopen(name, "rb")) == NULL)
	{fl_message("ERROR opening input file!\nClose the window to exit");exit(1);}
	
	fseek(in_fp, 0L, SEEK_END);	// go to file's end
	file_size = ftell(in_fp);
	fseek(in_fp, 0L, SEEK_SET);	// go to beginning	
	
   fread (&array, 3600, 1, in_fp);	// 1 object 3600 bytes long

	//   Convert Textual File Header:  EBCDIC-to-ASCII
	for(i=0;i<3200;i++) text_hdr[i]=to_ascii(array[i]);
	for(i=1;i<3200;i++) if(i%80==0) text_hdr[i-1]='\n';
		
   hdr2=(short*)array;
   format=hdr2[1612];				// Correspond to byte 3225
   
   if(format>0 && format<9) ; // Native format - do nothing
   else								// swap bytes (for both IBM and IEEE)
   {
		for(i=0;i<24;i++) convert2(array+3212+i*2);
      format=hdr2[1612];		
		if(format>0 && format<9) ;
		else 
		{
			fl_message("ERROR: input is not a SEG-Y file!\nClose the window to exit");
			exit(1);
		}
   }   
	
   dsamp=(int)hdr2[1608];	// SAMPLE_INTERVAL at byte 3217 <-- GLOBAL
   nsamp=(int)hdr2[1610];	// NUMBER_SAMPLES at byte 3221  <-- GLOBAL
   measure=hdr2[1627];		// Measurement system at byte 3255
	
   if(dsamp<0 || dsamp>8000 || nsamp<10 || nsamp>16000)
   { 
		fl_message("ERROR:  irrational sampling information.\n\
Sample Interval = %d microseconds\nNumber of Trace Samples = %d\n\
Close the window to exit...",dsamp,nsamp);	
		fclose(in_fp);   exit(1);
   }

	tr_size = 60*sizeof(int) + nsamp*sizeof(float);

	Ntr = (int)((file_size-3600L)/(nsamp*sizeof(float)+240));
	
   convert_mode=1;  // No conversion
   if(format==5) convert_mode=2; // IEEE-to-native
   if(format==1) convert_mode=3;	// IBM--to-native
	
	///////////////////////////////////////////////////////////
	/****       Initial Visualization Parameters          ****/
	
	tr_sta=1;
	if(Ntr<50) tr_fin=Ntr;
	else tr_fin=50;

	tr_vis = tr_fin-tr_sta+1; // if trace count starts from '1'
	samp_sta=0;  samp_fin=10;
	if(nsamp<100) samp_fin=10;
	else if(nsamp<500) samp_fin=100;
			else if(nsamp<1002) samp_fin=500;
					else samp_fin=1000;				
	samp_vis = samp_fin-samp_sta;
	gain = 1.0;
	
	/////////////////////////////////////////////////////////////
	/**      Get Normalization Factor for visualization       **/
	/**   by reading data until non-zero trace will be found  **/

	hdrs = new (nothrow) char[240];	
	seis = new (nothrow) float[nsamp];
	trace1 = new (nothrow) float[nsamp];	
	trace2 = new (nothrow) float[nsamp];
				
	Norm=1.0;
	for(tr=0;tr<Ntr;tr++)
	{
		offset=3600L+(long)tr_size*tr;	
		fseek(in_fp, offset, SEEK_SET);
		fread (hdrs, 240, 1, in_fp);
		if(convert_mode>1) convert_hdrs(hdrs);
		hdr2=(short*)hdrs; 
		if(hdr2[14]==2) continue;	// Trace ID code (0=Unknown,1=Seismic,2=Dead)
		fread (seis, nsamp*sizeof(float), 1, in_fp);
		if(convert_mode>1)
		{
			if(convert_mode==2)			// IEEE-to-native
				for(i=0;i<nsamp;i++) seis[i]=efloat(seis[i]);
			if(convert_mode==3)			// IBM-to-native
				for(i=0;i<nsamp;i++) seis[i]=ibm2ieee(seis[i]);			
		} 	
		
		count=avr=0.0;
		for(i=0;i<nsamp;i++) 
		{ 
			amp=fabs(seis[i]); 
			if(amp>0.0) { avr+=amp; count+=1.0; }
		}
		if(count>0.0) { avr/=count; Norm=4.0*avr; break; } 
	}
		
	if(hdr2[34] >= 0) elev_scal=(float)hdr2[34];
	else elev_scal=(float)(-1.0/hdr2[34]);
	if(hdr2[35] >= 0) coor_scal=(float)hdr2[35];
	else coor_scal=(float)(-1.0/hdr2[35]);	

}
////////////////////////////////////////////////////////////////
void read_seis_dat()		// fill up seis[] with data; float *seis;
{							// Do memory re-allocation before reading
	int i,tr;
	long offset;	
	
	delete[] hdrs; // originally allocated in read_file_hdr();	
	hdrs = new (nothrow) char[tr_vis * 240];		

	delete[] seis; // originally allocated in read_file_hdr();	
	seis = new (nothrow) float[(tr_vis+1) * nsamp];
	if(seis == 0)
	{ 
		fl_message("ERROR: NO memory to accomodate seismic data!\n\
Close the window to exit...");	
		fclose(in_fp);  exit(1);
	}
	
/****   seis[..] contains the whole trace from t=0 to 'nsamp'   ****/

	for(tr=0;tr<tr_vis;tr++)
	{	// tr_size = 60*sizeof(int) + nsamp*sizeof(float);
		offset = 3600L + (long)tr_size*(tr_sta+tr-1);	
		fseek(in_fp, offset, SEEK_SET);
		fread (hdrs+tr*240, 240, 1, in_fp); // 240 bytes-long object
      if(convert_mode>1) convert_hdrs(hdrs+tr*240);		
		// fseek(in_fp, samp_sta*sizeof(float), SEEK_CUR);		
		fread (seis+tr*nsamp, nsamp*sizeof(float), 1, in_fp);
		if(convert_mode>1)
		{
			if(convert_mode==2)			// IEEE-to-native
				for(i=0;i<nsamp;i++) 
					seis[tr*nsamp+i]=efloat(seis[tr*nsamp+i]);
			if(convert_mode==3)			// IBM-to-native
				for(i=0;i<nsamp;i++) 
					seis[tr*nsamp+i]=ibm2ieee(seis[tr*nsamp+i]);			
		} 		
	}
}
////////////////////////////////////////////////////////////////////
//	 Resampling using Cubic Convolution Method [R. Keys, 1981].
// Four data values from an array  dat[0..NS1]  are used every
// time to RETURN a calculated "index" value of the RESAMPLED
// array having size NS2 and same origin as the input array.
////////////////////////////////////////////////////////////////
float conv3(float *dat, int NS1, int NS2, int index)
{
	float y_1,y0,y1,y2,dx,dx2,dx3,k_1,k0,k1,k2;
	
	if(index<0 || index >=NS2) return 0.0;
	
	int i = index*NS1/NS2;
	
	y0=dat[i]; 	
	if(i<1) y_1=y0; else y_1=dat[i-1];
	if(i>NS1-2) y1=y0; else y1=dat[i+1]; 
	if(i>NS1-3) y2=y1; else y2=dat[i+2];
	
	dx = (float)(index*NS1 - i*NS2)/NS2;		
	dx2=dx*dx;  dx3=dx*dx2;		
	
	k_1 = (2.0*dx2 - dx3 - dx)/2.0;
	k0 = (3.0*dx3 - 5.0*dx2 + 2.0)/2.0;
	k1 = (4.0*dx2 - 3.0*dx3 + dx)/2.0;
	k2 = (dx3 - dx2)/2.0;
	return (k_1*y_1 + k0*y0 + k1*y1 + k2*y2);	
}
//
void convert_hdrs(char* array)  //  Byte-swapping for the
{                               //  240-byte array of trace headers
   int k;
   void convert2(char* input);
   void convert4(char* input);
	
   for(k=0;k<7;k++) convert4(array+k*4);       // Bytes   1 - 28
   for(k=0;k<4;k++) convert2(array+28+k*2);    // Bytes  29 - 36
   for(k=0;k<8;k++) convert4(array+36+k*4);    // Bytes  37 - 68
   for(k=0;k<2;k++) convert2(array+68+k*2);    // Bytes  69 - 72
   for(k=0;k<4;k++) convert4(array+72+k*4);    // Bytes  73 - 88
   for(k=0;k<46;k++) convert2(array+88+k*2);   // Bytes  89 - 180
   for(k=0;k<5;k++) convert4(array+180+k*4);   // Bytes 181 - 200
   for(k=0;k<2;k++) convert2(array+200+k*2);   // Bytes 201 - 204
   convert4(array+204);  convert2(array+208);  // Bytes 205 - 210
   for(k=0;k<5;k++) convert2(array+210+k*2);   // Bytes 211 - 224
   convert4(array+224);  convert2(array+228);  // Bytes 225 - 230
   convert2(array+230);                        // Bytes 231 - 233
	// Unassigned bytes 233 - 240
}
////////////////////////////////////////////////////////////////
//
// SUBROUTINES
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// http://pubs.usgs.gov/of/2005/1311/
//
// Adopted by Yuri Ganshin
//
void convert2(char* input) {
	
   unsigned char temp;
	
   temp = input[1];
   input[1] = input[0];
   input[0] = temp;
}

void convert4(char* input) {
	
   unsigned char temp;
	
   temp = input[3];
   input[3] = input[0];
   input[0] = temp;
	
   temp = input[2];
   input[2] = input[1];
   input[1] = temp;
}

float efloat(float value) {
	
   union {
      char bytes[4];
      float flt;
   } ieee;
	
	// swap bytes
   unsigned char temp;
   ieee.flt = value;
	
   temp = ieee.bytes[3];
   ieee.bytes[3] = ieee.bytes[0];
   ieee.bytes[0] = temp;
	
   temp = ieee.bytes[2];
   ieee.bytes[2] = ieee.bytes[1];
   ieee.bytes[1] = temp;

   return(ieee.flt);
}

float ibm2ieee(const float ibm) 
{
   long IntMantissa;
   float Mantissa,PosResult;
   // DEFINE AND SETUP UNION VARIABLES
   union { float f; unsigned char c[4]; } src; 
	
   // ASSIGN INPUT PARAMETER TO UNION MEMBER
   src.f = ibm; // When you change the value of a union member, 
	// you modify all the members of the union
   // CONVERTION
   IntMantissa = ((long)src.c[1] << 16)+((long)src.c[2]<<8)+src.c[3];
   Mantissa = (float)(IntMantissa)/(float)(0x1000000);
   PosResult = Mantissa*pow(16.0,(double)((src.c[0]&0x7F)-64));
	
   if (src.c[0] & 0x80) return -PosResult;
   else return PosResult;
}

/*
 **  ASCII <=> EBCDIC conversion functions
 */
unsigned char to_ascii ( unsigned char c )
{
	unsigned char e2a[256] = {
		0,  1,  2,  3,156,  9,134,127,151,141,142, 11, 12, 13, 14, 15,
		16, 17, 18, 19,157,133,  8,135, 24, 25,146,143, 28, 29, 30, 31,
		128,129,130,131,132, 10, 23, 27,136,137,138,139,140,  5,  6,  7,
		144,145, 22,147,148,149,150,  4,152,153,154,155, 20, 21,158, 26,
		32,160,161,162,163,164,165,166,167,168, 91, 46, 60, 40, 43, 33,
		38,169,170,171,172,173,174,175,176,177, 93, 36, 42, 41, 59, 94,
		45, 47,178,179,180,181,182,183,184,185,124, 44, 37, 95, 62, 63,
		186,187,188,189,190,191,192,193,194, 96, 58, 35, 64, 39, 61, 34,
		195, 97, 98, 99,100,101,102,103,104,105,196,197,198,199,200,201,
		202,106,107,108,109,110,111,112,113,114,203,204,205,206,207,208,
		209,126,115,116,117,118,119,120,121,122,210,211,212,213,214,215,
		216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,
		123, 65, 66, 67, 68, 69, 70, 71, 72, 73,232,233,234,235,236,237,
		125, 74, 75, 76, 77, 78, 79, 80, 81, 82,238,239,240,241,242,243,
		92,159, 83, 84, 85, 86, 87, 88, 89, 90,244,245,246,247,248,249,
		48, 49, 50, 51, 52, 53, 54, 55, 56, 57,250,251,252,253,254,255
	};
	
	return e2a[c];
}
